const utils = require('@alssoftware/alliance-npm-lambda')
const Promise = require('bluebird')
const moment = require('moment')
const uuidV1 = require('uuid/v4')

const config = require('./config.json')

const lambdaName = 'machineOrchestratorProgramSend'
const serializer = utils.Serializer
const postgres = new utils.Postgres(config.databaseName, config.databaseUser, config.databasePassword, config.databaseHost, config.databasePort, config.databaseLogging)
const sns = new utils.Sns(config.awsRegion, config.awsAccessKeyId, config.awsSecretAccessKey)
const s3 = new utils.S3(config.awsRegion, config.awsAccessKeyId, config.awsSecretAccessKey)
const machineCommand = new utils.MachineCommand(config.firebaseServiceAccount, config.firebaseUrl)
const firebaseMachineCommandLimit = config.firebaseMachineCommandLimit ? config.firebaseMachineCommandLimit : 5

exports.handler = (event, context) => {
  try {
    const reqObj = serializer.deserializeRequestObject('send-payload', { machineProgramId: 'machineProgramId' }, 'createdBy', event)
    const machinesSQL = `SELECT row_to_json(A) FROM (SELECT * FROM lms_machine.vw_machine WHERE id IN (${reqObj.machineIds.join()})) AS A;`
    const machineProgramSQL = `SELECT row_to_json(A) FROM (SELECT * FROM lms_machine.machine_program WHERE id = '${reqObj.machineProgramId}') AS A;`
    const roomMachineCommands = {}
    const createdAt = moment().format(utils.Constants.Reports.COMPAT_TIMESTAMP_FORMAT)

    Promise.all([
      postgres.get(null, null, machinesSQL),
      postgres.get(null, null, machineProgramSQL)
    ])
      .spread((machines, machineProgram) => {
        if (machines.length < 1) throw new Error(`No machines were found using machine ids:${reqObj.machineIds.join()}.`)
        if (machineProgram.length < 1) throw new Error(`Machine program with id ${reqObj.machineProgramId} was not found.`)
        else machineProgram = machineProgram[0]
        let roomCommandIndex = 1
        let prevMachineRoomId = 0
        // Note: A machine command needs to be created for each room.
        for (let i = 0, len = machines.length; i < len; i++) {
          if (prevMachineRoomId !== machines[i].room.id && i !== 0) { roomCommandIndex += 1 }
          if (roomMachineCommands[roomCommandIndex] == null) {
            roomMachineCommands[roomCommandIndex] = {
              id: uuidV1(),
              room: {
                id: machines[i].room.id
              },
              machines: [],
              params: machineProgram.settingBytes,
              type: 'PROGRAM_MACHINE',
              createdAt,
              createdBy: reqObj.createdBy
            }
          }

          roomMachineCommands[roomCommandIndex].machines.push({
            id: machines[i].id,
            controlId: machines[i].controlId
          })
          if (roomMachineCommands[roomCommandIndex].machines.length === firebaseMachineCommandLimit) { roomCommandIndex += 1 }
          prevMachineRoomId = machines[i].room.id
        }
        // console.log(JSON.stringify(roomMachineCommands));

        // Build the items needed to save to S3.
        const bucketName = config.roomMachineCommandsBucket
        const partitionDate = `partition_date=${moment(createdAt).format('YYYY-MM-DD')}`
        const s3Functions = []
        Object.keys(roomMachineCommands).forEach((indexId) => {
          const roomId = roomMachineCommands[indexId].room.id
          const partitionRoomId = `room_id=${roomId}`
          const fileName = `${roomId}_${moment(createdAt).format(utils.Constants.Reports.ISO_8601_TIMESTAMP_FORMAT)}_SERVER_GENERATED_${roomMachineCommands[indexId].id}.json`
          s3Functions.push(s3.save(bucketName, `${partitionDate}/${partitionRoomId}/${fileName}`,
            JSON.stringify(flattenRoomMachineCommand(roomMachineCommands[indexId]))))
        })
        return Promise.all(s3Functions)
      })
      .then(() => {
        const commandFunctions = []
        Object.keys(roomMachineCommands).forEach((indexId) => {
          commandFunctions.push(machineCommand.saveToFirebase(roomMachineCommands[indexId]))
        })
        return Promise.all(commandFunctions)
      })
      .then(() => {
        context.succeed()
      })
      .catch((error) => {
        sns.publishConflict(config.snsTargetArn, lambdaName, event, context, error)
      })
  } catch (error) {
    sns.publishConflict(config.snsTargetArn, lambdaName, event, context, error)
  }
}

let flattenRoomMachineCommand = (roomMachineCommand) => {
  const flattenedRoomMachineCommand = {
    event_uuid: roomMachineCommand.id,
    room_id: roomMachineCommand.room.id,
    machines: roomMachineCommand.machines,
    params: roomMachineCommand.params,
    type: roomMachineCommand.type,
    received_at: roomMachineCommand.createdAt,
    created_at: roomMachineCommand.createdAt,
    created_by: roomMachineCommand.createdBy.id
  }
  return flattenedRoomMachineCommand
}
